import java.awt.*;

/**
 * A World is made up of a 2-dimensional grid of Cells. Each Cell is square in
 * shape.
 *
 * @author Phil Reakes after Ryan Heise
 */
public class Cell
    extends Panel
{
    /**
     * A reference to the world in which this Cell exists.
     */
    private World world;

    /**
     * The location of this Cell in the world.
     */
    private Point position;
    
    /**
     * A reference to the mine on this Cell, if any.
     */
    private Mine mine = null;
    
    /**
     * A reference to the mine on this Cell, if any.
     */
    private Marker marker = null;        
        
    /**
     * Whether the Cell has a mine or not;
     */
    private boolean hasMine = false;
    
    /**
     * Whether the Cell has a marker or not;
     */
    private boolean hasMarker = false;  
    
    /**
     * Whether the Cell has been cleared or not;
     */
    private boolean isClear = false;     
    
    /**
     * Constructs a new Cell in the specified World.
     *
     * @param world the World this Cell will be a part of.
     * @param xpos the x coordinate of the Cell.
     * @param ypos the y coordinate of the Cell.
     */
    public Cell(World world, int xpos, int ypos)
    {
        this.world = world;
        position = new Point(xpos, ypos);
        
        setBackground(new Color(170, 170, 170));
        setLayout(new BorderLayout());
    }
    
    /**
     * Get a reference to the World containing this Cell.
     * 
     * @return  a reference to the world containing this Cell
     */
    public World getWorld()
    {
        return world;
    }       
    
    /**
     * Gets the location of this Cell.
     *
     * @return the location of this Cell.
     */
    public Point getPosition()
    {
        return position;
    }

    /**
     * Reveals whether the Cell contains a mine or not.
     * 
     * @return true if a mine is present, false otherwise
     */
    public boolean hasMine()
    {
        return hasMine;
    }
    
    /**
     * Adds a mine to the Cell.
     */
    public void addMine()
    {
        mine = new Mine( this );
        hasMine = true;
    }
    
    /**
     * Removes a mine from the Cell.
     */
    public void removeMine()
    {
        mine = null;
        hasMine = false;
    }
    
    /**
     * Get a reference to the mine on this Cell.
     * 
     * @return a reference to the mine on this Cell, or null if none present
     */
    public Mine getMine()
    {
        return mine;
    }    

    /**
     * Reveals whether the Cell contains a marker or not.
     * 
     * @return true if a marker is present, false otherwise
     */
    public boolean hasMarker()
    {
        return hasMarker;
    }
    
    /**
     * Adds a flag type marker to the Cell.
     */
    public void addMarker()
    {
        marker = new Marker( this );
        hasMarker = true;
    }
    
    /**
     * Removes a marker from the Cell.
     */
    public void removeMarker()
    {
        marker = null;
        hasMarker = false;
    }
    
    /**
     * Get a reference to the marker on this Cell.
     * 
     * @return a reference to the marker on this Cell, or null if none present
     */
    public Marker getMarker()
    {
        return marker;
    }
    
    /**
     * Get a reference to the marker on this Cell.
     * 
     * @return a reference to the marker on this Cell, or null if none present
     */
    public void cycleMarker()
    {
        if( hasMarker )
        {
            marker.cycle();
        }
        else
        {
            addMarker();
        }
    }       


    /**
     * Reveals whether the Cell is clear or not.
     * 
     * @return true if a cell is clear, false otherwise
     */
    public boolean isClear()
    {
        return isClear;
    }

    
    /**
     * Clears a non-mined Cell.
     */
    public void clear()
    {
        // set Cell's background colour to the dark grey
        // of the gridlines
        setBackground( new Color( 150, 150, 150 )  );
        isClear = true;
    }

    /**
     * Clear all adjacent cells if they are completely empty.
     * 
     * Beware -- contains recursion!
     */
    public void cascade()
    {
        // clear this cell
        clear();
                                
        // if any adjacent cells are marked, we can't clear them, so return.
        if( getAdjacentMarkers() > 0 ) return;
        
        // if all adjacent cells are empty, cascade them all
        if( getAdjacentMineNumbers() == 0 )
        {
            // clear all adjacent cells
                                            
            // check each neighbouring cell in turn
            for( int i = 0; i < 8; i++ )
            {
                Direction direction = new Direction( i * 45 );
                Cell neighbour = getNextCell( direction );
            
                // if the cell is on the world...
                if( neighbour != null )
                {
                    // ...and isn't already clear
                    if( !neighbour.isClear() ) 
                    {
                        // clear it using a cascade
                        neighbour.cascade();
                    }
                 }
             } // for
         } // if no adjacent mine numbers
         else // if mine numbers adjacent
         {
            // check each neighbouring cell in turn
            for( int i = 0; i < 8; i++ )
            {
                Direction direction = new Direction( i * 45 );
                Cell neighbour = getNextCell( direction );
            
                // if the cell is on the world...
                if( neighbour != null )
                {
                    // ...and isn't already clear
                    if( !neighbour.isClear() ) 
                    {
                        
                        if( neighbour.getAdjacentMines() > 0 )
                        {
                            // display its mine number colour
                            neighbour.paintMineNumberColour();
                        }
                        else
                        {
                            // or cascade it if none
                            neighbour.cascade();
                        }    
                    }
                 }
             } // for         
         
         }
    } // method cascade()
    
    /**
     *  Gets the number of mines adjacent to the Cell.
     *  
     *  @return     the number of mines adjacent to the Cell.
     */
    public int getAdjacentMines()
    {
        int mineCount = 0;
        
        // check each neighbouring cell in turn
        for(int i = 0; i < 8; i++)
        {
            Direction direction = new Direction( i * 45 );
            Cell cell = getNextCell( direction );
            
            // if the cell is on the world...
            if( cell != null )
            {
                // ...and it has a mine on it
                if( cell.hasMine() ) 
                {
                    mineCount++;
                }
            }
        }
        
        return mineCount;
    }
    
    /**
     *  Gets the number of markers adjacent to the Cell.
     *  
     *  @return     the number of markers adjacent to the Cell.
     */
    public int getAdjacentMarkers()
    {
        int markerCount = 0;
        
        // check each neighbouring cell in turn
        for(int i = 0; i < 8; i++)
        {
            Direction direction = new Direction( i * 45 );
            Cell cell = getNextCell( direction );
            
            // if the cell is on the world...
            if( cell != null )
            {
                // ...and it has a marker on it
                if( cell.hasMarker() ) 
                {
                    markerCount++;
                }
            }
        }
        
        return markerCount;
    } 
    
    /**
     *  Gets the number of 'mine numbers' adjacent to the Cell.
     *  
     *  @return     the number of maine numbers adjacent to the Cell.
     */
    public int getAdjacentMineNumbers()
    {
        int mineNumberCount = 0;
        
        // check each neighbouring cell in turn
        for(int i = 0; i < 8; i++)
        {
            Direction direction = new Direction( i * 45 );
            Cell cell = getNextCell( direction );
            
            // if the cell is on the world...
            if( cell != null )
            {
                // ...and it has a mine number on it
                if( cell.getAdjacentMines() > 0 ) 
                {
                    mineNumberCount++;
                }
            }
        }
        
        return mineNumberCount;
    }     
    
    /**
     * Paints a Cell's 'mine number colour' on that Cell. The
     * 'mine number colour' differs depending on the
     * the number of mines adjacent to the Cell. As a mnemonic
     * it follows the colour spectrum from lowest
     * wavelength to highest.
     * <p>
     * Key to 'mine colours':
     * <p>
     * Orange -     1 mine adjacent.
     * Yellow -     2 mines adjacent.
     * Green -      3 mines adjacent.
     * Cyan -       4 mines adjacent.
     * Blue -       5 mines adjacent.
     * Magenta -    6 mines adjacent.
     * Dark Grey -  7 mines adjacent.
     * Black -      8 mines adjacent.
     */
    public void paintMineNumberColour()
    {
        int numberOfMines = getAdjacentMines();
        
        Color color = Color.black;
        
        switch( numberOfMines )
        {
            case 1:
                color = Color.orange;
                break;
            case 2:
                color = Color.yellow;
                break;
            case 3:
                color = Color.green;
                break;
            case 4:
                color = Color.cyan;
                break;
            case 5:
                color = Color.blue;
                break;
            case 6:
                color = Color.magenta;
                break;
            case 7:
                color = Color.darkGray;
                break;                
        }
        
        setBackground( color );
        
        // don't let the cell be clicked again
        isClear = true;        
    }    
       
    /**
     * Gets the reference to a neighbouring Cell in the specified direction.
     *
     * @param direction the direction in which to grab the neighboring Cell.
     * The Direction will be automatically rounded to the nearest Cell.
     * @return the neighbouring Cell in the specified direction.
     */
    public Cell getNextCell(Direction direction)
    {
        double radians = direction.getRadians();
        int relativeX = (int)Math.round(Math.cos(radians));
        int relativeY = (int)Math.round(Math.sin(radians));
        int targetX = position.x + relativeX;
        int targetY = position.y - relativeY; // XXX: which way's up?
        
        Cell target = world.getCell(targetX, targetY);
        if (target == null)
            return null;
               
        return target;
    }
    
    /**
     * This method lets the layout manager know what size this Cell would
     * prefer to be.
     *
     * @return the preferred size.
     */
    public Dimension getPreferredSize()
    {
        return new Dimension(20, 20);
    }
         
    /**
     * Creates a String representation of this Cell. This is useful for
     * debugging purposes.
     *
     * @return a String representation of this Cell.
     */
    public String toString()
    {
        return "cell["+position.x+","+position.y+"]";
    }
}
