import java.awt.Color;

/**
 * A Marker is a type of CellDweller. A marker is a flag or question mark 
 * placed by the player on a particular Cell. There can only be one Marker 
 * on each Cell at a time. 
 * 
 * @author Phil Reakes
 */
public class Marker extends CellDweller
{

    // colour representing a cell flagged by the player
    //  as definitely containing a mine
    final static Color FLAG = Color.pink;

    // colour representing a cell flagged by the player
    //  as possibly containing a mine    
    final static Color QUESTION_MARK = Color.white;
    
    // whether the marker is a flag or question mark
    private Color markerType;

    /**
     * Constructor for Markers. Places the marker on a specified Cell.
     * When constructed, a Marker defaults to the flag type.
     * 
     * @param   cell  the Cell on which this marker is to be placed.
     */
    public Marker( Cell cell )
    {   
        // call CellDweller's constructor
        super( cell );
        
        // set all new Markers to flag type
        setType( Marker.FLAG );
        
        // update flag counters
    }
       
    /**
     * Get the Color representing the 'type' of this 
     * marker -- flag or question mark.
     * 
     * @return     the 'type' of the marker
     */
    public Color getType()
    {
        return markerType;
    }
    
    /**
     * Set the 'type' of this marker -- flag or question mark,
     * and paint its colour on the parent cell.
     * 
     * @param   type    the 'type' of the marker
     */
    public void setType( Color type )
    {
        markerType = type;
        parentCell.setBackground( type );        
    }        

    /**
     * Update World's flag counters to keep score accurate.
     */
    public void updateFlagCounters()
    {
        World   world =             parentCell.getWorld();
        int     correctFlags =      world.getFlaggedMines();
        int     incorrectFlags =    world.getFlaggedNonMines();
             
        if( markerType == Marker.FLAG )
        {
            if( parentCell.hasMine() )
            {
                // add a mine correctly flagged
                world.setFlaggedMines( correctFlags++ );
            }
            else
            {
                // add a mine incorrectly flagged
                world.setFlaggedNonMines( incorrectFlags++ );
            }
        } // if
        else // if marker is questionmark
        {
            if( parentCell.hasMine() )
            {
                // subtract a mine correctly flagged
                world.setFlaggedMines( correctFlags-- );
            }
            else
            {
                // subtract a mine incorrectly flagged
                world.setFlaggedNonMines( incorrectFlags-- );
            }        
        } // else
    } // method updateFlagCounters()
    
    /**
     * Cycles the marker between flag, question mark, and nothing.
     */
    public void cycle()
    {    
        // if marker is a flag, set it to a question mark
        if( markerType == Marker.FLAG )
        {
            setType( Marker.QUESTION_MARK );
            
            // update flag counters
        }
        else
        {
            // else if it's a question mark, delete it
            parentCell.setBackground( new Color( 170, 170, 170 ) );            
            parentCell.removeMarker();

        }
    }  // method cycle()
}
