import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.util.List;
import java.lang.reflect.*;

/**
 * A World is made up from a grid of square-shaped Cells. 
 * 
 * @author Phil Reakes after Ryan Heise
 */
public class World
    extends Panel
{
    /**
     * This random generator is used by all worlds (yes, it is possible to
     * create more than one world!), particularly in picking a random cell to
     * place new things when they are added to the world.
     */
    private static Random random = new Random();
    
    private Panel cellsPanel;
    //private ControllersPanel controllersPanel;
    
    /**
     * The number of columns in this world.
     */
    private int colCount;

    /**
     * The number of rows in this world.
     */
    private int rowCount;
    
    /**
     * The number of mines in this world.
     */
    private int mineCount;  
    
    /**
     * The number of correctly flagged mines in this world.
     */
    private int flaggedMines = 0;     
    
    /**
     * The number of flagged non-mine cells in this world.
     */
    private int flaggedNonMines = 0;      
    
    /**
     * The number of cells the player has clicked so far.
     */
    private int cellClicks = 0;
    
    /**
     * Whether game is in progress or over.
     */
    private boolean gameOver = false; 
    
    /**
     * A reference to the WorldFrame containing this World. 
     */
    private WorldFrame worldFrame;
    
    /**
     * A 2-dimensional array of cells in this world.
     */
    private Cell[][] cells;
    private CellListener cellListener;
        
    /**
     * Creates a new world with the specified number of columns and rows.
     * Unlike the previous constructor, this one will not automatically create
     * a new window. The world object created by this constructor is simply a
     * GUI component that can be added to some other window. You should create
     * your own window and add this GUI component to it.
     *
     * @param colCount the number of columns in the world.
     * @param rowCount the number of rows in the world.
     * @param frame a window that you have created to contain this world.
     */
    public World(int colCount, int rowCount, WorldFrame frame)
    {
        this.colCount = colCount;
        this.rowCount = rowCount;
        
        worldFrame = frame;
        
        // Determine the number of mines in the world, e.g. in a 15x15 world, 10 mines
        mineCount = (int) Math.ceil( (colCount + rowCount) / 2 ) + 2;
        
        // figure out where to put the mines
                
        setBackground(new Color(150,150,150));

        setLayout(new BorderLayout());
        cellsPanel = new Panel();
        add(cellsPanel, BorderLayout.CENTER);

        cells = new Cell[colCount][rowCount];

        cellsPanel.setLayout(new WorldLayout(colCount, rowCount));

        cellListener = new CellListener();
        for (int y = 0; y < rowCount; y++)
        {
            for (int x = 0; x < colCount; x++)
            {
                Cell cell = new Cell(this, x, y);
                cells[x][y] = cell;
                cellsPanel.add(cell, WorldLayout.CELL);
                cell.addMouseListener(cellListener);
            }
        }
 
        // add mines to world
        int minesLeft = mineCount;
                
        // distribute mines
        while( minesLeft > 0 )
        {
            // get a random cell
            int x = random.nextInt(colCount);
            int y = random.nextInt(rowCount);
            Cell cell = cells[x][y];
            
            // if the cell doesn't have a mine under it...
            if ( cell.hasMine() == false )
            {
                  // ...add a mine
                  cell.addMine();
                  minesLeft--;
            }
            
            // until all mines are placed
        }
        
    }

    /**
     * Returns the number of cells the player has clicked.
     */
    public int getCellClicks()
    {
        return cellClicks;
    }

    /**
     * Sets the number of cells the player has clicked. Not
     * used at the moment but included for completeness.
     * 
     * @param   clicks    the desired number of clicks
     */    
    public void setCellClicks(int clicks)
    {
        cellClicks = clicks;
    }

    /**
     * Reveals whether the game is over.
     */    
    public boolean gameIsOver()
    {
        if( gameOver ) 
            return true;
        else
            return false;
    }
    
    /**
     * Sets the game over state to true or false.
     * 
     * @param   gameState   true if setting game over, false otherwise
     */
    public void setGameOver( boolean gameState )
    {
        gameOver = gameState;
        
        if( gameOver )
        {
            displayScore();
        }
    }  

    /**
     * Get the number of mines that have been correctly flagged.
     * 
     * @return  the number of mines correctly flagged
     */
    public int getFlaggedMines()
    {  
        return flaggedMines;
    }
    

    /**
     * Set the number of mines that have been correctly flagged.
     * 
     * @param  the number of mines correctly flagged
     */
    public void setFlaggedMines( int mines )
    {  
        flaggedMines = mines;
    }    

    /**
     * Get the number of cells that have been incorrectly flagged.
     * 
     * @return  the number of cells incorrectly flagged
     */
    public int getFlaggedNonMines()
    {  
        return flaggedNonMines;
    }
    

    /**
     * Set the number of cells that have been incorrectly flagged.
     * 
     * @param  the number of cells incorrectly flagged
     */
    public void setFlaggedNonMines( int cells )
    {  
        flaggedNonMines = cells;
    }  

    /**
     * Displays the player's score. 
     */
    public void displayScore()
    {
        String  scoreMessage;
        int     score;
    
        // check whether player has won
        
        // loop through every cell to check for incorrect flags
        for (int y = 0; y < rowCount; y++)
        {
            for (int x = 0; x < colCount; x++)
            {
                Cell cell = getCell( x, y );
                
                // if cell has no mine, check whether it was flagged 
                if( cell.hasMarker() && !cell.hasMine() )
                {
                    Marker marker = cell.getMarker();
                    // penalise player for incorrect flagging
                    if( marker.getType() == Marker.FLAG )
                    {
                        flaggedNonMines++;
                    }
                } // if cell has marker
                
            } // for
        }  // for      
        
        /* If all mines have been correctly flagged and no cells have been 
         * incorrectly flagged, player wins. Score is in this case is:
         *
         * 100 * ( 1000 * mines flagged ) all over number of clicks
         * 
         * This is so that the score gets higher the faster you complete
         * the game.
         */
        if( ( flaggedMines == mineCount ) && ( flaggedNonMines == 0 ) )
        {
            scoreMessage = "You win!";
            score = (int) 100 * ( ( 1000 * flaggedMines ) / getCellClicks() );
        }
        else // if player loses
        {
            // reveal unflagged mines
            
            // loop through every cell to check for missing flags
            for (int y = 0; y < rowCount; y++)
            {
                for (int x = 0; x < colCount; x++)
                {
                    Cell cell = getCell( x, y );
                
                    // if cell has mine, check whether it was flagged 
                    if( cell.hasMine() )
                    {
                        Mine mine = cell.getMine();

                        if( !mine.isFlagged() )
                        {
                            mine.reveal();
                        }
                    } // if cell has mine
                } // for
            }  // for               
        
            scoreMessage = "You lost.";        
            // reward player for correcly flagging mines and staying alive            
            score = (int) ( 5 * ( flaggedMines - flaggedNonMines ) ) + getCellClicks();
        }
        
        // display the player's score
        scoreMessage += " Score: " + score + ".";
        worldFrame.updateScoreCard( scoreMessage );
        
    } // method calculateScore()
 
    /**
     * Gets the cell at coordinates (x,y). The origin (0,0) is in the upper
     * left corner of the world.
     *
     * @param x the x coordinate
     * @param y the y coordinate
     * @return the cell at position (x,y).
     */
    Cell getCell(int x, int y)
    {
        try
        {
            return cells[x][y];
        }
        catch (ArrayIndexOutOfBoundsException e)
        {
            return null;
        }
    }
    

    class CellListener
        extends MouseAdapter
    {
        public void mouseClicked(MouseEvent event)
        {
            Cell cell = (Cell) event.getSource();
            
            int mods = event.getModifiers();

            // only handle Cell clicks and keep score if a game is being played
            // and the cell has not been cleared already
            if( !gameIsOver() && !cell.isClear() )
            {                         
                // handle left mouse clicks
                if( (mods & event.BUTTON1_MASK) != 0)
                {     
                      // if the player clicks a live mine...
                      if( cell.hasMine() ) 
                      {                  
                            // calculate score and end game
                            setGameOver( true );
                            
                      } // if cell has mine
                      else
                      {                                                      
                            int adjacentMines = cell.getAdjacentMines();
                    
                            // if the cell has mines adjacent
                            if( adjacentMines > 0 )
                            {
                                cell.paintMineNumberColour();
                            }
                            else
                            {
                                cell.cascade();
                            }
                            
                     } // else if cell doesn't have a mine

                    // maybe access through a method?
                    cellClicks++;
            
                } // end if left button clicked
                else
                {
                    // if right or centre button clicked, handle markers
                    cell.cycleMarker();
                                      
                    // if the cell has a mine, check if it was properly flagged
                    if( cell.hasMine() )
                    {
                        Mine mine = cell.getMine();
                    
                        if( mine.isFlagged() )
                        {
                            // mine correctly flagged
                            flaggedMines++;
                        }
                     } // if cell has mine                    
                     
                     // if all mines now properly flagged
                     if( ( flaggedMines == mineCount ) && ( flaggedNonMines == 0 ) )
                     {
                        setGameOver( true );
                     }
                    
                } // else right button clicked
            } // if gameIsOver()
        } // method mouseClicked
        
    }

    /**
     * The world is just a GUI component that needs to be added to a window.
     * By default, a DefaultFrame will be created to contain the world.
     */
    class DefaultFrame
        extends Frame
    {
        DefaultFrame()
        {
            setLayout(new BorderLayout());
            add(World.this, BorderLayout.CENTER);
            
            addWindowListener(new WindowAdapter()
                    {
                        public void windowClosing(WindowEvent event)
                        {
                            dispose();
                            System.exit(0);
                        }
                    });
            
            pack();
        }
    }

    class WorldLayout
        implements LayoutManager2
    {
        public static final String CELL = "Cell";
        private int colCount;
        private int rowCount;

        private int cellSize = 20;
        private int gapSize = 1;
        private Dimension preferredSize;

        private Cell[][] cells;

        private int cellCount = 0;
        private int hgapCount = 0;
        private int vgapCount = 0;
    
        public WorldLayout(int colCount, int rowCount)
        {
            this.colCount = colCount;
            this.rowCount = rowCount;

            cells = new Cell[colCount][rowCount];

            preferredSize = new Dimension(cellSize*colCount + gapSize*(colCount+1), cellSize*rowCount + gapSize*(rowCount+1));
        }

        private void addCell(Cell cell)
        {
            int x = cellCount % colCount;
            int y = cellCount / colCount;
            cells[x][y] = cell;
            cellCount++;
        }

        public void addLayoutComponent(Component comp, Object constraints)
        {
            addLayoutComponent((String)constraints, comp);
        }
        
        public void addLayoutComponent(String name, Component comp)
        {
            synchronized (comp.getTreeLock())
            {
                if (CELL.equals(name))
                    addCell((Cell)comp);
                else
                    throw new IllegalArgumentException("Unknown constraint: " + name);
            }
        }
        
        public void removeLayoutComponent(Component comp) {
            throw new RuntimeException("NOT IMPLEMENTED");
        }
        
        public Dimension minimumLayoutSize(Container target)
        {
            return new Dimension(0, 0);
        }
        
        public Dimension preferredLayoutSize(Container target)
        {
            return preferredSize;
        }
        
        public Dimension maximumLayoutSize(Container target)
        {
            return new Dimension(Integer.MAX_VALUE, Integer.MAX_VALUE);
        }
        
        public float getLayoutAlignmentX(Container parent)
        {
            return 0.5f;
        }
        
        public float getLayoutAlignmentY(Container parent)
        {
            return 0.5f;
        }
        
        public void invalidateLayout(Container target)
        {
        }
        
        public void layoutContainer(Container target)
        {
            synchronized (target.getTreeLock())
            {
                Insets insets = target.getInsets();
                int top = insets.top;
                int bottom = target.getHeight() - insets.bottom;
                int left = insets.left;
                int right = target.getWidth() - insets.right;
    
                Component c = null;

                for (int row = 0; row < rowCount; row++)
                {
                    for (int col = 0; col < colCount; col++)
                    {
                        Cell cell = cells[col][row];
                        int x = left + gapSize*(col+1) + cellSize*col;
                        int y = top + gapSize*(row+1) + cellSize*row;
                        cell.setBounds(x, y, cellSize, cellSize);
                    }
                }

            }
        }
        public String toString()
        {
            return getClass().getName();
        }
    }
}
